#include <iostream>
#include <cmath>
#include <cstring>
#include "Analysis.h"

#define DEFAULT_NOISE 3.0
#define DEFAULT_PED 512.0

#define INIT_ARRAY(V) memset((void *)(V), 0, sizeof(V))

Analysis::Analysis() :
    ConfigFile("analysis"), _ped_weight(0.01), _noise_weight(0.001),
    _seedcut(5.), _neighcut(3.), _sncut(5.), _mask(0),
    _do_cmmd(true), _update(true)
{
   INIT_ARRAY(_ped);
   INIT_ARRAY(_noise);
   INIT_ARRAY(_hitmap);
   INIT_ARRAY(_signal);
   INIT_ARRAY(_raw);
   INIT_ARRAY(_sn);
   INIT_ARRAY(_header);
   INIT_ARRAY(_cmmd);
   INIT_ARRAY(_cmmd_noise);

}

Analysis::~Analysis()
{

}

void Analysis::cpy(const Analysis &a)
{
    reset();
    cpy_data(a);
}

void Analysis::cpy_data(const Analysis &a)
{
    memcpy(_ped, a._ped, sizeof(_ped));
    memcpy(_noise, a._noise, sizeof(_noise));
    memcpy(_hitmap, a._hitmap, sizeof(_hitmap));
    memcpy(_signal, a._signal, sizeof(_signal));
    memcpy(_raw, a._raw, sizeof(_raw));
    memcpy(_sn, a._sn, sizeof(_sn));
    _ped_weight = a._ped_weight;
    _noise_weight = a._noise_weight;
    _seedcut = a._seedcut;
    _neighcut = a._neighcut;
    _sncut = a._sncut;
    _update = a._update;
    _do_cmmd = a._do_cmmd;
}

void Analysis::reset(bool keep_pedestals)
{
    int i;
    if (keep_pedestals)
    {
        for (i=0; i<256; i++)
        {
            _hitmap[i] = 0.;
        }
    }
    else
    {
        for (i=0; i<256; i++)
        {
            _ped[i] = DEFAULT_PED;
            _noise[i] = DEFAULT_NOISE;
            _hitmap[i] = 0.0;
            _raw[i] = 0.0;
            _signal[i] = 0.0;
        }
    }
}

void Analysis::process_event(const unsigned short *v, int mask)
{
    int i, j;
    int nzero=0;
    int nchan, nchip = 0;
    int coff = 0;

    _mask = mask;
    if ( (mask & 3) == 3)
    {
        nchip = 2;
        nchan = 256;
        for (int ichip=0, i=0; ichip<2; ichip++)
        {
            const unsigned short *d = v + (ichip*144 + 16);
            for (j=0; j<128;j++, i++)
            {
                _raw[i] = double(d[j]);
            }
        }
    }
    else
    {
        coff = (mask&1) ? 0 : 128;
        for (i=0; i<256; i++)
            _raw[i] = 0.;

        for (i=0; i<128; i++)
            _raw[coff+i] = double(v[16+i]);
    }

    
    // clear list of hits
    clear();
    
//    for (i=0; i<256; i++)
//        _raw[i] = double(v[i]);

    for (i=0; i<256; i++)
    {
        int ichip = i/128;
        if ( !(mask & 1<<ichip) )
        {
            signal(i, 0.);
            sn(i, 0);
            continue;
        }
        if (noise(i)<=1.)
        {
            noise(i, DEFAULT_NOISE);
            nzero++;
        }
        signal(i, _raw[i] - ped(i));
        sn(i, fabs(signal(i) / noise(i)));
    }

    for (i=0; i<2; i++)
    {
        _cmmd[i] = 0.;
        _cmmd_noise[i] = DEFAULT_NOISE;
    }
    if (do_cmmd() )
    {
        if (nzero<128)
            common_mode();
    }

    if (_update)
        update_pedestals(_raw);

    find_clusters();
}

void Analysis::update_pedestals(const double *v)
{
    int i;
    for (i=0; i<256; i++)
    {
        int ichip = i/128;
        if ( !(_mask & 1<<ichip) )
            continue;

        // IF noise is 0, set it arbitrarily to 1.
        if (noise(i)==0.)
            noise(i, 1.);

        if (ped(i)==0.)
        {
            // If pedestal is not yet computed we assume the current
            // channel value should not be too far
            ped(i, v[i]);
        }
        else
        {
            // Do the pedestal and noise correction
            double corr;
            double xs;

            signal(i, v[i] - ped(i));
            corr = signal(i) * ped_weight();

            signal(i, v[i]- ped(i) - _cmmd[i/128]);
            xs = signal(i) / noise(i);
            sn(i, fabs(xs));
            if (corr > 1.)
                corr = 1.;

            if (corr < -1)
                corr = -1.;

            ped(i, ped(i) + corr);

            if (fabs(xs) < _sncut)
            {
                noise(i, noise(i)*(1. - noise_weight()) + xs * xs * noise_weight());
            }
        }
    }
}

void Analysis::find_clusters()
{
    // This is someting simple...
    // Just take the highest
    int i, imax=-1;
    double mxsig=-1.e20, sg, val;

    clear();

    for (i=0; i<256; i++)
    {
        if (sn(i)>_seedcut)
        {
            val = fabs(signal(i));
            if (mxsig<val)
            {
                mxsig = val;
                imax = i;
            }
        }
    }

    if (imax<0)
        return;

    sg = signal(imax);

    // Now look at neighbours
    if ( (i=imax-1)>=0)
    {
        if (fabs(sn(i))>_neighcut)
            sg += signal(i);
    }

    // Now look at neighbours
    if ( (i=imax+1)<256)
    {
        if (fabs(sn(i))>_neighcut)
            sg += signal(i);
    }

    add_hit(imax, sg);
}

void Analysis::common_mode()
{
    int ip, i, j;

    
    for (j=0; j<2; j++)
    {
        if (!(_mask & 1<<j) )
            continue;

        int offs = 128*j;
        int sstop = 128*(j+1);
        double mean, sm, xn, xx, xs, xm, tmp;
        bool use_it;
        mean = sm = 0.;
        //std::cout << "Chip " << j << std::endl;    
        for (ip=0;ip<3;ip++)
        {
            xn = xs = xm = 0.;
            for (i=offs; i<sstop; i++)
            {
                use_it = true;
                xx = _raw[i]-_ped[i];
                if (ip)
                {
                    tmp = fabs((xx-mean)/sm);
                    use_it = (tmp<2.5);
                }
                if (use_it)
                {
                    xn++;
                    xm += xx;
                    xs += xx * xx;
                }
            }
            if (xn>0.)
            {
                mean = xm / xn;
                sm = sqrt( xs/xn - mean*mean);
            }
            //std::cout << "...iter " << ip << ": xm " << mean << " xs: " << sm << std::endl;
        }
        _cmmd[j] = mean;
        _cmmd_noise[j] = sm;
        for (i=0; i<128; i++)
            _signal[i+offs] = _raw[i+offs]-_ped[i+offs]-mean;
    }
}

void Analysis::load_data()
{
    if (exists() )
    {
        int i, length;
        double *v;
        _do_cmmd = get_bool("common_mode");
        _ped_weight = get_double("ped_weight");
        _noise_weight = get_double("noise_weight");

        v = get_double_list("pedestals", length);
        for (i=0; i<length; i++)
            ped(i, v[i]);

        delete [] v;

        v = get_double_list("noise", length);
        for (i=0; i<length; i++)
            noise(i, v[i]);

        delete [] v;

        _seedcut = get_double("seed_cut");
        _neighcut = get_double("neigh_cut");
        _sncut = get_double("sn_cut");
    }
}

void Analysis::save_data()
{
    set_bool("common_mode", _do_cmmd);
    set_double("seed_cut", _seedcut);
    set_double("neigh_cut", _neighcut);
    set_double("sn_cut", _sncut);
    set_double("ped_weight", _ped_weight);
    set_double("noise_weight", _noise_weight);
    set_double_list("pedestals", pedestals(), 256);
    set_double_list("noise", noise(), 256);
}
