#ifndef __Alibava_Fifo_h__
#define __Alibava_Fifo_h__
#include <glib.h>

#if HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#endif

#if HAVE_GIO
#include <gio/gio.h>
#endif

class USBport;

class Fifo
{
    private:
        GCond      *cond;
        GMutex     *cond_mutex;
        GMutex     *data_mutex;
        GThread    *thread;
        GIOChannel *chan;
#ifdef HAVE_GIO
        GSocket    *sckt; // TODO: convert into a GSource
        GCancellable *cancellable;
#endif
        int         fd;
        bool        in_loop;
        bool        _end_poll;
        /// buffer size
        unsigned int bfsiz;

        /// buffer
        char *buff;

        /// read pointer
        char *rdptr;

        /// write pointer
        char *wrptr;

        /// End of the buffer
        char *end;

        USBport *_P;

        void notify();
        void grow(int sz);
        void read_from_channel();

        static gboolean data_from_channel(GIOChannel *source, GIOCondition condition, Fifo *);
        static gpointer start_loop(Fifo *);
#ifdef G_OS_WIN32
        static gpointer start_win_loop(Fifo *);
#endif
        int from_file(void *buf, int size);
        int from_source(void *buf, int size);

    public:
        Fifo(unsigned int size);
        virtual ~Fifo();

        /// returns the size of the fifo
        int size() const
        {
            return bfsiz;
        }
        
        /// resizes the fifo
        void resize(unsigned int sz);

        /// tells if the fifo is empty
        bool empty()
        {
            return rdptr == wrptr;
        }

        /// reads from the fifo
        int read(char *buf, int);
        
        /** writes to the fifo
         If force is false, the data will be lost when there is not enough
         space in the buffer. If force is true, the Fifo buffer will be
         enlarged to ensure that this block of data is not lost
         */
        int write(char *buf, int, bool force = false);
        
        int flush();

        /// resets the fifo
        void reset()
        {
            rdptr = wrptr = buff;
        }
        
        /// Returns the space left in the FIFO buffer
        int space_left()
        {
            int left;
            if (rdptr == wrptr)
                left = bfsiz;
            else
            {
                left = ((rdptr + bfsiz - wrptr) % bfsiz);
            }
            return left;
        }

        
        
        /// Returns the ammount of data not yet read from the FIFO
        int data_available()
        {
            int left;
            if (wrptr > rdptr) // no wrap yet
                left = wrptr - rdptr;
            else
            {
                left = (end - rdptr) + (wrptr - buff);
            }
            return left;
        }
        
        /**
         * waits until there is data available
         * returns true if there is data and false on timeout
         */
        bool wait(double tmout=0.);
        
        /**
         * Attach the fifo to an open file
         */
        int attach(int fd);

#ifdef G_OS_WIN32
        /**
         * This is for windows serial ports
         */
        int attach(USBport *);
#endif

        /*
         * TODO: we have to change to a "generic" GSource
         */
#if HAVE_GIO
        int attach(GSocket *S, GIOChannel *);
#endif
        void detach();
        bool running() const { return in_loop; }
        bool endpoll() const { return _end_poll; }
        void endpoll(bool x);
 
};

#endif /*__Alibava_Fifo_h__*/
