/*
 * path.cc
 *
 *  Created on: 27 May 2011
 *      Author: lacasta
 */

#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include "path_utils.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#ifndef _WIN32
#include <pwd.h>
#endif
#include <limits>


#ifdef _WIN32
#define DIR_SEPARATOR '\\'
#define DIR_SEPARATOR_S "\\"
#define SEARCHPATH_SEPARATOR ';'
#define SEARCHPATH_SEPARATOR_S ";"
#else
#define DIR_SEPARATOR '/'
#define DIR_SEPARATOR_S "/"
#define SEARCHPATH_SEPARATOR ':'
#define SEARCHPATH_SEPARATOR_S ":"
#endif


/**
 * Splits a string into tokens separated by any of the given
 * separators.
 */
template <typename Container>
void stringtok (Container &container, std::string const &in,
                const char * const delimiters = " \t\n")
{
    const std::string::size_type len = in.length();
    std::string::size_type i = 0;

    while ( i < len )
    {
        // eat leading whitespace
        i = in.find_first_not_of (delimiters, i);
        if (i == std::string::npos)
            return;   // nothing left but white space

        // find the end of the token
        std::string::size_type j = in.find_first_of (delimiters, i);

        // push token
        if (j == std::string::npos)
        {
            container.push_back (in.substr(i));
            return;
        }
        else
            container.push_back (in.substr(i, j - i));

        // set up for next loop
        i = j + 1;
    }
}


bool PathUtils::isabs(const std::string &s)
{
	if (isalpha(s[0]) && s[1]==':' && s[2]==DIR_SEPARATOR)
		return true;

    return s[0] == DIR_SEPARATOR;
}

bool PathUtils::exists(const std::string &s)
{
    struct stat sb;
    if ( stat(s.c_str(), &sb) == -1 )
        return false;
    else
        return true;
}

std::string PathUtils::dir_name( const std::string& path )
{
    return path.substr( 0, path.find_last_of( DIR_SEPARATOR ) +1 );
}

std::string PathUtils::base_name( const std::string& path )
{
    return path.substr( path.find_last_of( DIR_SEPARATOR ) +1 );
}

std::string PathUtils::join(const std::string a, std::string b)
{
    std::string the_path = a + DIR_SEPARATOR_S + b;
    std::string out = abspath(the_path);
    return out;
}


std::string PathUtils::join(const std::string a, std::vector< std::string > p)
{
    std::string path = a;
    std::vector< std::string >::iterator ip;
    for (ip = p.begin(); ip!=p.end(); ip++)
    {
        if ( (*ip)[0] == DIR_SEPARATOR)
        {
            path = *ip;
        }
        else if ( path.empty() || path[ path.length()-1] == DIR_SEPARATOR)
        {
            path += *ip;
        }
        else
        {
            path += DIR_SEPARATOR + *ip;
        }
    }

    return path;
}


std::string PathUtils::expanduser(const std::string &in_path)
{
#ifdef _WIN32
	return in_path;
#else
    std::string path = in_path;
    if ( path[0] != '~' )
        return path;

    std::string userhome;
    std::string::size_type ipos = path.find(DIR_SEPARATOR,1);
    if (ipos == path.npos)
        ipos = path.length();

    if (ipos==1)
    {
        char *home = getenv("HOME");
        if (home==0)
        {
            userhome = getpwuid(getuid())->pw_dir;
        }
        else
            userhome = home;
    }
    else
    {
        struct passwd *pwent;
        pwent = getpwnam( path.substr(1, ipos).c_str() );
        if (!pwent)
            return path;
        else
            userhome = pwent->pw_dir;
    }

    return userhome + path.substr(ipos);
#endif
}

std::string PathUtils::normpath(const std::string &in_path)
{
    std::string path = in_path;
    if (path.empty())
        return ".";

    int length = path.length();
    int initial_slashes = (path[0] == DIR_SEPARATOR ? 1 : 0);
    if ( length > 1 && initial_slashes && path[1] == DIR_SEPARATOR)
    {
        if (length<3 || path[2]!=DIR_SEPARATOR)
            initial_slashes = 2;
    }

    std::vector< std::string > comps, new_comps;
    std::vector< std::string >::iterator comp;
    stringtok(comps, path, DIR_SEPARATOR_S );

    for (comp = comps.begin(); comp != comps.end(); comp++)
    {
        if ( (*comp).empty() || (*comp)==".")
            continue;

        if (  *comp != ".." ||
              ( initial_slashes==0 && new_comps.empty() ) ||
              ( !new_comps.empty() && new_comps[new_comps.size()-1]==".." ) )
        {
            new_comps.push_back(*comp);
        }
        else
        {
            new_comps.pop_back();
        }
    }
    path.clear();
    int i=0, n = new_comps.size()-1;
    for (comp=new_comps.begin(); comp!=new_comps.end(); i++, comp++)
    {
        path +=  *comp;
        if (i<n)
            path += DIR_SEPARATOR_S;
    }
    if (initial_slashes)
    {
        std::string slashes = std::string(initial_slashes,DIR_SEPARATOR);
        path = slashes + path;
    }
    if (path.empty())
        path = ".";

    return path;
}

std::string PathUtils::getcwd()
{
    char line[1024];
    std::string cwd = ::getcwd(line, sizeof(line));
    return cwd;
}

std::string PathUtils::abspath(const std::string &in_path)
{
    std::string path = in_path;
    if ( ! isabs(path) )
    {
        path = getcwd() + DIR_SEPARATOR_S + path;
    }
    return normpath(path);
}

PathUtils::FileParts PathUtils::splitext(const std::string &path)
{
    std::string out = normpath(path);
    std::string::size_type dpos = out.find_last_of( '.' );
    return FileParts(out.substr( 0, dpos), out.substr(dpos) );
}

std::string PathUtils::find_bin(const std::string &path)
{
    if ( exists(path) )
        return path;

    std::string bname = base_name(path);
    std::string PATH( getenv("PATH") );

    std::vector< std::string > dpath;
    std::vector< std::string >::iterator ipath;
    stringtok(dpath, PATH, SEARCHPATH_SEPARATOR_S);
    for (ipath=dpath.begin(); ipath!=dpath.end(); ipath++)
    {
        std::string p = *ipath + DIR_SEPARATOR_S + bname;
        if ( exists(p) )
            return p;
    }

    return bname;
}

#define MAX_DIR_PATH 2048
void PathUtils::find_file(const std::string pattern, std::vector<std::string> &out, std::string cwd, bool recurse)
{
    DIR* dir; /* pointer to the scanned directory. */
    struct dirent* entry; /* pointer to one directory entry.   */
    struct stat dir_stat; /* used by stat().                   */

    if (cwd.empty())
        cwd=".";

    /* open the directory for reading */
    dir = opendir(cwd.c_str());
    if (!dir)
    {
        fprintf(stderr, "Cannot read directory '%s': ", cwd.c_str());
        perror("");
        return;
    }

    /* scan the directory, traversing each sub-directory, and */
    /* matching the pattern for each file name.               */
    while ((entry = readdir(dir)))
    {
        /* check if the pattern matchs. */
        std::string dname = entry->d_name;
        if ( dname=="." || dname=="..")
            continue;

        std::string fname = join(cwd, dname);
        /* check if the given entry is a directory. */
        if (stat(fname.c_str(), &dir_stat) == -1)
        {
            perror("stat:");
            continue;
        }

        if ( recurse && S_ISDIR(dir_stat.st_mode))
        {
            find_file(pattern, out, join(cwd, fname), recurse);
            fprintf(stderr, "PathUtils::find_file with recursion not implemented.\n");
            continue;
        }

        if ( fname.find(pattern) != fname.npos)
        {
            out.push_back( fname );
        }
    }
}
